# Snowpark Container Service GPU instance type and corresponding GPU counts.
INSTANCE_TYPE_TO_GPU_COUNT = {
    "GPU_3": 1,
    "GPU_5": 1,
    "GPU_7": 4,
    "GPU_10": 8,
    "GPU_NV_S": 1,
    "GPU_NV_M": 4,
    "GPU_NV_L": 8,
}
